********** CP/M DISK I/O ROUTINES  **********
**********			   **********
VERS	EQU	'3b'		; February 15,1983 15:32 klf "M320.ASM"
*********************************************
	DW	MODLEN,BUFLEN

	TITLE	'SASI- DRIVER FOR MMS CP/M SASI BUS INTERFACE'
;			PRELIMINARY RELEASE FOR XEBEC/SEAGATE
	MACLIB	Z80
	$-MACRO

BASE	EQU	0000H	;ORG FOR RELOC

***** PHYSICAL DRIVES ARE ASSIGNED AS FOLLOWS  *****
*****					       *****
*****  50-58 RESERVED FOR SASI BUS DISKS       *****
*****					       *****
****************************************************

PAGE


***************************************************
**  PORTS AND CONSTANTS
***************************************************
GPIO	EQU	0F2H		; SWITCH 501

ACK	EQU	00000001B
INT	EQU	00000010B
PER	EQU	00000100B
BUSY	EQU	00001000B
CMND	EQU	00010000B
MSG	EQU	00100000B
POUT	EQU	01000000B
REQ	EQU	10000000B

RUN	EQU	00000000B
SWRS	EQU	00010000B
INTE	EQU	00100000B
SEL	EQU	01000000B

RECAL	EQU	1	; COMMAND OP CODES
RDBL	EQU	8
WRBL	EQU	10
INIT	EQU	12

DPHDPB	EQU	10
DPHL	EQU	16
DPBL	EQU	21
CSTRNG	EQU	13
NPART	EQU	19
CBYTE	EQU	4
DDATA	EQU	5
DCTYPE	EQU	3
SECTBL	EQU	20
DDPB	EQU	47

WRALL	EQU	0	; WRITE TO ALLOCATED
WRDIR	EQU	1	; WRITE TO DIRECTORY
WRUNA	EQU	2	; WRITE TO UNALLOCATED
READOP	EQU	3	; READ OPERATION

***************************************************
** LINKS TO REST OF SYSTEM
***************************************************

PATCH	EQU	BASE+1600H
MBASE	EQU	BASE		; MODULE BASE
COMBUF	EQU	BASE+0C000H	; COMMON BUFFER
BUFFER	EQU	BASE+0F000H	; MODULE BUFFER

***************************************************
** PAGE ZERO ASSIGNMENTS
***************************************************
	ORG	0
?CPM		DS	3
?DEV$STAT	DS	1
?LOGIN$DSK	DS	1
?BDOS		DS	3
?RST1		DS	3
?CLOCK		DS	2
?INT$BYTE	DS	1
?CTL$BYTE	DS	1
		DS	77
?FCB		DS	36
?DMA		DS	128
?TPA		DS	0
***************************************************
PAGE



***************************************************
** OVERLAY MODULE INFORMATION ON BIOS
***************************************************

	ORG	PATCH
	DS	51		; BIOS JUMP TABLE

DISK$STAT	DB	0

		DS	8

DRIV0	EQU	50
MIXER:	DB	50,51,52	; PHYSICAL DRIVES 50,51
	DS	13		; LOGICAL-PHYSICAL DRIVE TABLE

DRIVE$BASE:
	DB	50,53		; DRIVE MODULE BASE TABLE
	DW	MBASE
	DS	28

CBIOS:	DS	3

NEWBAS	DS	2
NEWDSK	DS	1
NEWTRK	DS	1
NEWSEC	DS	1
HRDTRK	DS	2
DMAA	DS	2

***************************************************
PAGE



***************************************************
** START OF RELOCATABLE DISK I/O MODULE
*************************************************** 

	ORG	MBASE		; START OF MODULE

	JMP	SEL$SASI
	JMP	READ$SASI
	JMP	WRITE$SASI

;	TEXT
	DB	'77320 SASI Interface (2 partitions) v2.24'
	DW	VERS,'$'


DPH0:	DW	0,0,0,0,DIRBUF,DPB0,CSV0,ALV0
DPH1:	DW	0,0,0,0,DIRBUF,DPB1,CSV0,ALV1
DPH2:	DW	0,0,0,0,DIRBUF,DPB2,CSV0,ALV2
DPH3:	DW	0,0,0,0,DIRBUF,DPB3,CSV0,ALV3
DPH4:	DW	0,0,0,0,DIRBUF,DPB4,CSV0,ALV4
DPH5:	DW	0,0,0,0,DIRBUF,DPB5,CSV0,ALV5
DPH6:	DW	0,0,0,0,DIRBUF,DPB6,CSV0,ALV6
DPH7:	DW	0,0,0,0,DIRBUF,DPB7,CSV0,ALV7
DPH8:	DW	0,0,0,0,DIRBUF,DPB8,CSV0,ALV8

CNUM:	DB	0		; CONTROLLER NUMBER 0

;	SECTOR DEFINITION/TRANSLATION TABLE
;		--ADDRESS--,   FLAG BYTE
DDEFTBL:DB	0,   0,   0,   0
	DB	0,   0,   0,   1
	DB	0,   0,   0,   2
	DB	0,   0,   0,   0
	DB	0,   0,   0,   0
	DB	0,   0,   0,   0
	DB	0,   0,   0,   0
	DB	0,   0,   0,   0
	DB	0,   0,   0,   0

;
;
; DISK PARAMETER BLOCKS -- CONTAIN DUMMY DATA. REAL DATA IS OBTAINED FROM 
;			   MAGIC SECTOR ON INITIALIZATION OF PARTITION
;
DPB0:	DW	64		; SPT
	DB	5,31,1		; BSH,BLM,EXM
	DW	648-1,512-1	; DSM,DRM
	DB	0F0H,0		; AL0,AL1
	DW	0,2		; CKS,OFF
	DB	00000010B,10000000B,0  ; MODE BYTES
	DB	0FFH,0FFH,0FFH	; MODE MASKS

DPB1:	DW	64		; SPT
	DB	5,31,1		; BSH,BLM,EXM
	DW	649-1,512-1	; DSM,DRM
	DB	0F0H,0		; AL0,AL1
	DW	0,2		; CKS,OFF
	DB	00000010B,10000000B,0  ; MODE BYTES
	DB	0FFH,0FFH,0FFH	; MODE MASKS

DPB2:	DW	64		; SPT
	DB	5,31,1		; BSH,BLM,EXM
	DW	607-1,512-1	; DSM,DRM
	DB	0F0H,0		; AL0,AL1
	DW	0,2		; CKS,OFF
	DB	00000010B,10000000B,0  ; MODE BYTES
	DB	0FFH,0FFH,0FFH	; MODE MASKS

DPB3:	DW	64		; SPT
	DB	5,31,1		; BSH,BLM,EXM
	DW	607-1,512-1	; DSM,DRM
	DB	0F0H,0		; AL0,AL1
	DW	0,2		; CKS,OFF
	DB	00000010B,10000000B,0  ; MODE BYTES
	DB	0FFH,0FFH,0FFH	; MODE MASKS

DPB4:	DW	64		; SPT
	DB	5,31,1		; BSH,BLM,EXM
	DW	607-1,512-1	; DSM,DRM
	DB	0F0H,0		; AL0,AL1
	DW	0,2		; CKS,OFF
	DB	00000010B,10000000B,0  ; MODE BYTES
	DB	0FFH,0FFH,0FFH	; MODE MASKS

DPB5:	DW	64		; SPT
	DB	5,31,1		; BSH,BLM,EXM
	DW	607-1,512-1	; DSM,DRM
	DB	0F0H,0		; AL0,AL1
	DW	0,2		; CKS,OFF
	DB	00000010B,10000000B,0  ; MODE BYTES
	DB	0FFH,0FFH,0FFH	; MODE MASKS

DPB6:	DW	64		; SPT
	DB	5,31,1		; BSH,BLM,EXM
	DW	607-1,512-1	; DSM,DRM
	DB	0F0H,0		; AL0,AL1
	DW	0,2		; CKS,OFF
	DB	00000010B,10000000B,0  ; MODE BYTES
	DB	0FFH,0FFH,0FFH	; MODE MASKS

DPB7:	DW	64		; SPT
	DB	5,31,1		; BSH,BLM,EXM
	DW	607-1,512-1	; DSM,DRM
	DB	0F0H,0		; AL0,AL1
	DW	0,2		; CKS,OFF
	DB	00000010B,10000000B,0  ; MODE BYTES
	DB	0FFH,0FFH,0FFH	; MODE MASKS

DPB8:	DW	64		; SPT
	DB	5,31,1		; BSH,BLM,EXM
	DW	607-1,512-1	; DSM,DRM
	DB	0F0H,0		; AL0,AL1
	DW	0,2		; CKS,OFF
	DB	00000010B,10000000B,0  ; MODE BYTES
	DB	0FFH,0FFH,0FFH	; MODE MASKS
PAGE



;	SELECT DISK CODE
;
SEL$SASI:
	XRA	A
	STA	SELERR		; NO SELECT ERRORS (YET)
	CALL	INIT$SASI
	JNZ	ERREXT
	LDA	PNDWRT		; CLEAR ANY PENDING WRITE
	ORA	A
	CNZ	WR$SEC
	LDA	NEWDSK		; CALCULATE DRIVE NUMBER RELATIVE TO 0
	SUI	DRIV0
	STA	RELDSK
	LXI	H,DDEFTBL
	MOV	E,A		; NOW POINT TO THE CORRECT
	MVI	D,0		; ENTRY IN THE SECTOR
	DAD	D		; OFFSET TABLE
	DAD	D
	DAD	D
	DAD	D
	SHLD	SECPTR
	MOV	A,M
	ANI	11100000B
	STA	LUN		; SAVE LOGICAL UNIT NUMBER BITS
	MOV	B,E
	INR	B
	LXI	H,DPH0-DPHL	; POINT TO DPH TABLE
	LXI	D,DPHL		; LENGTH OF DPH's
SEL1:	DAD	D
	DJNZ	SEL1		; CALCULATE POINTER TO REQUESTED DPH
	LXI	D,DPHDPB
	DAD	D		; POINT TO ADDRESS OF DPB
	CALL	HLIHL		; DPB ADDRESS IN HL
	SHLD	CURDPB		; SAVE IT

	LXI	H,LUNDATA-16	; CALCULATE POINTER TO
	LXI	D,16		;  DATA FOR THIS LOGICAL UNIT
	LDA	LUN
	RLC
	RLC
	RLC
NXTLUN: DAD	D
	DCR	A
	JP	NXTLUN
	SHLD	DATAPTR

	CALL	CHK$INIT	; INITIALIZE DRIVE IF NECESSARY
	LHLD	CURDPB
	INX	H
	INX	H
	INX	H
	MOV	A,M		; GET BLOCK MASK
	STA	BLKMSK		; USED FOR UNALLOCATED RECORD COUNT
	LXI	D,10
	DAD	D
	MOV	A,M		; GET TRACK OFFSET
	STA	OFFSET
	INX	H
	INX	H		; POINT TO MODE BYTE 1
	MOV	A,M
	ANI	3		; ISOLATE SECTOR SIZE BITS
	STA	BLCODE		; SAVE AS DEBLOCKING CODE
RETSEL: LDA	RELDSK
	MOV	C,A
	LXI	D,DPH0
	RET

ERREXT: MVI	A,1
	STA	SELERR
	JR	RETSEL


;
;	DRIVER INITIALIZATION CODE
;
INIT$SASI:
	LDA	INIT$FLAG	; NEED TO INITIALIZE ?
	ORA	A
	RZ
	MVI	B,7CH
	IN	GPIO		; READ SWITCH 501
	ANI	00000011B	; WHAT'S PORT 7C SET FOR ?
	CPI	00000010B	;  IF Z67, THEN THIS IS IT
	JRZ	GOTPRT
	MVI	B,78H
	IN	GPIO		; READ SWITCH 501
	ANI	00001100B	; WHAT'S PORT 78 SET FOR ?
	CPI	00001000B	;  IF Z67, THEN THIS IS IT
	RNZ
GOTPRT: MOV	A,B
	STA	BASE$PORT	; SAVE BASE PORT ADDRESS
	XRA	A
	STA	INIT$FLAG	; FLAG DRIVER AS INITIALIZED
	RET

CHK$INIT:
	LHLD	SECPTR
	INX	H
	INX	H
	INX	H		; POINT TO FLAG BYTE
	MOV	A,M
	STA	FLAGS		; SAVE FOR USE BY INIT$DRIVE
	BIT	6,M		; IS IT A FLOPPY DISK ?
	STC
	RNZ			;  THEN DON'T INITIALIZE
	PUSH	H		; SAVE FLAG BYTE
	CALL	LOGIN		; HAS DISK BEEN LOGGED IN ?
	POP	H		; RECALL FLAG BYTE
	RC			;  THEN DON'T INITIALIZE
	BIT	5,M		; IS IT REMOVABLE MEDIA ?
	JRNZ	INIT$HARD	;  MUST INITIALIZE
	BIT	7,M		; CHECK INITIALIZATION BIT
	JRZ	INIT$HARD
	CALL	INIT$DRIVE	; ALWAYS SEND DATA TO CONTROLLER ON LOGIN
	JRNZ	INIT$ERR
	RET

INIT$HARD:
	MVI	A,RDBL		; OP CODE TO READ A SECTOR
	STA	CMBFR
	XRA	A		; SECTOR 0
	STA	CMBFR+1
	STA	CMBFR+2
	STA	CMBFR+3
	INR	A
	STA	CMBFR+4 	; READ 1 SECTOR
	CALL	RESET		; RESET CONTROLLER
	CALL	GETCON		; WAKE UP CONTROLLER
	CZ	OUTCOM		; OUTPUT READ COMMAND
	CZ	SASIRW		; READ IN SECTOR
	CZ	CHK$STAT	; CHECK STATUS OF READ
	JRNZ	INIT$ERR

	LDA	FLAGS
	ANI	00001111B	; ISOLATE PARTITION NUMBER BITS
	LXI	H,HSTBUF+NPART
	CMP	M		; ERROR IF PARTITION NUMBER OUT OF RANGE
	JRNC	INIT$ERR
	INR	A
	MOV	B,A		; GET DPB FROM 'MAGIC SECTOR'
	LXI	H,HSTBUF+DDPB-DPBL
	LXI	D,DPBL
NXTDPB: DAD	D
	DJNZ	NXTDPB
	LDED	CURDPB		; MOVE IT INTO DPB TABLE IN DRIVER
	LXI	B,21
	LDIR
	LXI	H,HSTBUF+DCTYPE ; GET L.U.N. SPECIFIC DATA FROM
	LDED	DATAPTR 	;  'MAGIC SECTOR'
	LXI	B,16		;  PUT IT INTO SPACE RESERVED FOR THIS L.U.N.
	LDIR
	MOV	B,A
	LXI	H,HSTBUF+SECTBL-3 ; POINT TO ENTRY IN SECTOR TABLE FROM
NXTPRT: INX	H		  ;  'MAGIC SECTOR'
	INX	H
	INX	H
	DJNZ	NXTPRT
	LDED	SECPTR		; POINT TO ENTRY IN DRIVE DEFINITION TABLE
	LDA	LUN
	ORA	M		; COMBINE L.U.N. AND MSB'S OF SECTOR NUMBER
	MOV	M,A
	LXI	B,3		; INSTALL SECTOR NUMBER 
	LDIR
	PUSH	D		; SAVE ADDRESS OF FLAG BYTE
	CALL	INIT$DRIVE	; SET INITIALIZATION BIT
	POP	H		; RECALL ADDRESS OF FLAG BYTE
	JRNZ	INIT$ERR	; ERROR ON PHYSICAL INITIALIZATION
	SETB	7,M		; SET INITIALIZATION BIT
	RET

INIT$ERR:
	POP	D		; CLEAR STACK
	JMP	ERREXT

LOGIN:	LDA	NEWDSK		; CHECK CP/M LOGIN VECTOR
	LXI	B,17
	LXI	H,MIXER
	CCIR
	MVI	A,17
	SUB	C
	MOV	B,A
	LXI	H,PATCH
	LXI	D,0D89H
	ORA	A
	DSBC	D
	CALL	HLIHL
	INX	H
	CALL	HLIHL
	CALL	HLIHL
ROTHL:	RARR	H
	RARR	L
	DJNZ	ROTHL
	RET

PAGE



READ$SASI:
	LDA	PNDWRT		; SECTOR WAITING TO BE WRITTEN ?
	ORA	A
	CNZ	WR$SEC
	MVI	A,READOP	; FLAG A READ OPERATION
	JR	RWOPER

WRITE$SASI:
	MOV	A,C

RWOPER: STA	WRTYPE		; SAVE WRITE TYPE
	LDA	SELERR
	ORA	A
	RNZ
	LDA	NEWDSK
	STA	REQDSK
	LHLD	HRDTRK
	SHLD	REQTRK
	MVI	C,0		; CALCULATE PHYSICAL SECTOR
	LDA	BLCODE		; PHYSICAL SECTOR SIZE CODE
	ORA	A		; TEST FOR ZERO
	MOV	B,A
	LDA	NEWSEC
	STA	REQSEC		; INITIAL GUESS IS 128 BYTE SECTORS
	JRZ	DBLOK3		; 128 BYTE SECTORS ?
DBLOK1: SRLR	A		; DIVIDE ACCUMULATOR BY 2
	RARR	C		; SAVE OVERFLOW BITS
	DJNZ	DBLOK1		; AND CONTINUE IF BLOCKING STILL <> 0
	STA	REQSEC		; SAVE IT
	LDA	BLCODE		; CALCULATE BLKSEC
	MOV	B,A		; FOR LOOPING
DBLOK2: RLCR	C		; NOE RESTORE THE OVERFLOW BY
	DJNZ	DBLOK2		; ROTATING IT RIGHT
DBLOK3: MOV	A,C
	STA	BLKSEC		; STORE IT
	MVI	A,0FFH
	STA	RD$FLAG 	; FLAG A PRE-READ
	LDA	WRTYPE
	RAR			; CARRY IS SET ON WRDIR AND READOP
	JRC	ALLOC		; NO NEED TO CHECK FOR UNALLOCATED RECORDS
	RAR			; CARRY IS SET ON WRUNA
	JRNC	CHKUNA
	SDED	URECORD 	; SET UNALLOCATED RECORD #
	DCR	A
	STA	UNALLOC 	; FLAG WRITING OF AN UNALLOCATED BLOCK
CHKUNA: LDA	UNALLOC 	; ARE WE WRITING AN UNALLOCATED BLOCK ?
	ORA	A
	JRZ	ALLOC
	LHLD	URECORD 	; IS REQUESTED RECORD SAME AS EXPECTED
	DSBC	D		;  SAME AS EXPECTED UNALLOCATED RECORD ?
	JRNZ	ALLOC		; IF NOT, THEN DONE WITH UNALLOCATED BLOCK
	XRA	A		; CLEAR PRE-READ FLAG
	STA	RD$FLAG
	INX	D		; INCREMENT TO NEXT EXPECTED UNALLOCATED RECORD
	SDED	URECORD
	LDA	BLKMSK
	ANA	E		; IS IT THE START OF A NEW BLOCK ?
	JRNZ	CHKSEC
ALLOC:	XRA	A		; NO LONGER WRITING AN UNALLOCATED BLOCK
	STA	UNALLOC

;*****************************************************************************
; CHKSEC: THIS SUBROUTINE COMPARES THE REQUESTED DISK TRACK AND SECTOR	     ;
;	  TO THE DISK,TRACK AND SECTOR CURRENTLY IN THE BUFFER. 	     ;
;	  OUTPUT: ZERO FLAG SET IF SAME, RESET IF DIFFERENT		     ;
;									     ;
CHKSEC: ANA	A		; CLEAR CARRY FOR DSBC			     ;
	LHLD	REQTRK							     ;
	LDED	OFFSET							     ;
	DSBC	D							     ;
	JRNZ	CHKBUF							     ;
	LDA	NEWSEC							     ;
	ORA	A		; FIRST SECTOR OF DIRECTORY ?		     ;
	JRZ	READIT							     ;
CHKBUF: LXI	H,REQDSK						     ;
	LXI	D,HSTDSK						     ;
	MVI	B,4							     ;
CHKBUF1:LDAX	D							     ;
	CMP	M							     ;
	JRNZ	READIT							     ;
	INX	H							     ;
	INX	D							     ;
	DJNZ	CHKBUF1 						     ;
	JR	NOREAD		;  THEN NO NEED TO PRE-READ		     ;
;*****************************************************************************

READIT: LDA	PNDWRT		; IS THERE A SECTOR THAT NEEDS TO BE WRITTEN ?
	ORA	A
	CNZ	WR$SEC		; WRITE IT
	LXI	D,HSTDSK	; SET UP NEW BUFFER PARAMETERS
	LXI	H,REQDSK
	LXI	B,4
	LDIR
	LDA	RD$FLAG 	; DO WE NEED TO PRE-READ ?
	ORA	A
	CNZ	RD$SEC		; READ THE SECTOR
NOREAD: LXI	H,HSTBUF	; POINT TO START OF SECTOR BUFFER
	LXI	B,128
	LDA	BLKSEC		; POINT TO LOCATION OF CORRECT LOGICAL SECTOR
MOVIT1: DCR	A
	JM	MOVIT2
	DAD	B
	JR	MOVIT1
MOVIT2: LDED	DMAA		; POINT TO DMA
	LDA	WRTYPE		; IS IT A READ OR A WRITE
	CPI	READOP
	JRZ	MOVIT3
	XCHG			; SWITCH DIRECTION OF MOVE FOR WRITE
	MVI	A,1		; FLAG A PENDING WRITE
	STA	PNDWRT
MOVIT3: LDIR			; MOVE IT
	LDA	WRTYPE		; CHECK FOR DIRECTORY WRITE
	DCR	A
	CZ	WR$SEC		; WRITE THE SECTOR IF IT IS
	XRA	A		; FLAG NO ERROR
	RET			; RETURN TO BDOS

WR$SEC: CALL	WRITE		; WRITE A PHYSICAL SECTOR
	MVI	A,0
	STA	PNDWRT		; FLAG NO PENDING WRITE
	RZ			; RETURN IF WRITE WAS SUCCESSFUL
	LDA	WRTYPE
	CPI	READOP		; IGNORE ERROR IF THIS IS A READ OPERATION
	RZ
	JR	RWERR

RD$SEC: CALL	READ		; READ A PHYSICAL SECTOR
	RZ			; RETURN IF SUCCESSFUL
	MVI	A,0FFH		; FLAG BUFFER AS UNKNOWN
	STA	HSTDSK
RWERR:	POP	D		; THROW AWAY TOP OF STACK
	MVI	A,1		; SIGNAL ERROR TO BDOS
	ORA	A
	RET			; RETURN TO BDOS



;
;	READ - WRITE ROUTINES
;
;	READ A PHYSICAL SECTOR CODE
;
READ:
	MVI	A,RDBL		; READ COMMAND CODE
	JR	DO$RW		; COMMON READ-WRITE ROUTINE
;
;	WRITE A PHYSICAL SECTOR CODE
;
WRITE:
	MVI	A,WRBL		; WRITE COMMAND CODE
;
;	COMMON READ-WRITE CODE
;
DO$RW:	STA	CMBFR		; COMMAND BUFFER OP CODE
	CALL	SET$SEC 	; CALCULATE AND INSTALL ACTUAL SECTOR
	CALL	WAKE$UP 	; SETUP CONTROLLER
	CZ	OUTCOM		; AND OUTPUT THE COMMAND
	CZ	SASIRW		; DO READ OR WRITE
	CZ	CHK$STAT	; CHECK THE BUS RESPONSE
	RET

RESET:	LDA	BASE$PORT
	INR	A
	MOV	C,A		; CONTROL PORT TO REG. C
	MVI	A,RUN		; CLEARS ANY PREVIOUS CONTROLLER ACTIVITY
	OUTP	A
	MVI	A,SWRS
	OUTP	A		; RESET CONTROLLER
	RET

INIT$DRIVE:
	CALL	RESET
	LHLD	DATAPTR 	; SEE IF IT'S XEBEC
	MOV	A,M
	ANI	11100000B
	JRNZ	NOTXBC		; SKIP IF NOT
	CALL	GETCON		; GET CONTROLLER'S ATTENTION
	LXI	H,ICMND 	; INITIALIZATION COMMAND STRING
	CZ	OUTCM0		; OUTPUT COMMAND
	RNZ
	LHLD	DATAPTR 	; DRIVE CHARACTERISTIC DATA
	INX	H
	INX	H
	MVI	B,8		; 8 BYTES LONG
	MVI	E,(REQ OR POUT OR BUSY)
	CALL	OUTCM1		; OUTPUT THE DATA
	CZ	CHK$STAT	;  AND CHECK STATUS
NOTXBC: LHLD	DATAPTR
	LXI	D,10		; NOW DO "ASSIGN DRIVE TYPE" COMMAND
	DAD	D		;  ( FOR DATA PERIPHERALS DONTROLLERS )
	PUSH	H
	CALL	GETCON		; GET CONTROLLER'S ATTENTION
	POP	H
	CZ	OUTCM0		; SEND THE COMMAND
	CZ	CHK$STAT
	RNZ
	CALL	GETCON		; NOW RESTORE THE HEADS
	LXI	H,RCMND
	CZ	OUTCM0
	CZ	CHK$STAT
	RET

PAGE



;	CALCULATE THE REQUESTED SECTOR
;
SET$SEC:
	LHLD	HSTTRK		; GET REQUESTED TRACK
	DAD	H		; *2
	DAD	H		; *4
	DAD	H		; *8
	DAD	H		; *16
	DAD	H		; *32
	DAD	H		; *64 (64 SECTORS/TRACK)
	PUSH	H
	LHLD	SECPTR
	MOV	C,M
	INX	H
	MOV	D,M
	INX	H
	MOV	E,M
	POP	H
	DAD	D		; ADD IN PARTITION OFFSET
	JRNC	NOCAR0		; CARRY FROM DAD (IF ANY) GOES INTO
	INR	C		;  HIGH ORDER BYTE OF SECTOR NUMBER
NOCAR0: LDA	BLCODE
	MOV	B,A
	ORA	A
	MOV	A,C
	JRZ	NODIV
	ANI	00011111B	; ELIMINATE L.U.N. BITS FROM DIVISION
	MOV	C,A
NXDIV:	SRAR	C
	RARR	H
	RARR	L
	DJNZ	NXDIV
NODIV:	LDA	HSTSEC		; GET REQUESTED SECTOR
	MOV	E,A
	MVI	D,0
	DAD	D		; ADD IT IN
	JRNC	NOCAR1
	INR	C
NOCAR1: MOV	A,C
	STA	CMBFR+1 	; MOVE TO COMMAND BUFFER
	MOV	A,H
	STA	CMBFR+2 	; MOVE REST OF SECTOR NUMBER TO COMMAND BUFFER
	MOV	A,L
	STA	CMBFR+3
	MVI	A,1		; TRANSFER 1 SECTOR
	STA	CMBFR+4
	LHLD	DATAPTR 	; GET CONTROL BYTE
	INX	H
	MOV	A,M
	STA	CMBFR+5 	; PUT INTO COMMAND BUFFER
	RET
PAGE

HLIHL:	MOV	A,M
	INX	H
	MOV	H,M
	MOV	L,A
	RET

WAKE$UP:CALL	GETCON
	RZ
	CALL	INIT$DRIVE
	RNZ
	CALL	GETCON
	RET


;
;	GET THE BUS' ATTENTION
;
GETCON:
	LDA	BASE$PORT
	MOV	C,A
	INR	C		; CONTROL PORT ADDRESS TO REG. C
	MVI	B,0		; TIMER COUNTER
GETCN1: DCR	C
	XRA	A
	OUTP	A		; CLEAR DATA REGISTER
	LDA	CNUM		; GET CONTROLLER NUMBER
	INR	C		; AND SEND IT TO THE CARD
	INR	C		; -SENC- PORT
	OUTP	A
	DCR	C		; CONTROL PORT
	INP	A		; READ CONTROL PORT
	ANI	BUSY
	JRZ	GETCN2
	DJNZ	GETCN1
	DCR	B		; RESET PSW/Z TO INDICATE ERROR
	RET
GETCN2: 
	MVI	A,SEL
	OUTP	A		; WAKE UP CONTROLER
	MVI	B,0
GETCN3:
	INP	A
	ANI	BUSY
	JRNZ	GETCN4
	DJNZ	GETCN3
	DCR	B		; RESET PSW/Z TO INDICATE ERROR
	RET
GETCN4: 
	MVI	A,RUN
	OUTP	A
	XRA	A		; NO ERROR
	RET

PAGE



;
; OUTCOM: OUTPUT A COMMAND TO THE DRIVE
;
OUTCOM: LXI	H,CMBFR
OUTCM0: LDA	LUN		; OR L.U.N. BITS INTO COMMAND
	INX	H
	ORA	M
	MOV	M,A
	DCX	H
	MVI	B,6		; COMMAND IS 6 BYTES LONG
	LDA	BASE$PORT
	MOV	C,A		; DATA PORT TO REG. C
	INR	A
	MOV	D,A		; CONTROL PORT TO REG. D
	MVI	E,(REQ OR CMND OR POUT OR BUSY)
OUTCM1: PUSH	B
	MVI	B,16		; SET LOOP COUNTER
	MOV	C,D		; CONTROL PORT ADDRESS TO REG. C
OUTLOP: INP	A
	ANI	(REQ OR CMND OR POUT OR BUSY)
	CMP	E
	JRZ	OUTOK
	DJNZ	OUTLOP
	DCR	B
	POP	B
	RET
OUTOK:	POP	B		; RETURNS DATA PORT ADDRESS TO REG. C
	OUTI			; OUTPUT COMMAND BYTE
	JNZ	OUTCM1
	XRA	A
	RET
;
;	ACTUAL READ-WRITE OF DATA
;
SASIRW: 			; THIS ROUTINE IS FOR READING AND WRITING
	LDA	CMBFR
	SUI	RDBL		; IS COMMAND A READ ?
	MVI	A,0B2H		; INIR FOR READS
	JRZ	NREAD
	MVI	A,0B3H		; OUTIR FOR WRITES
NREAD:	STA	HERE+1
	LXI	H,HSTBUF	; AND WRITING DATA
	LDA	BASE$PORT
	MOV	C,A		; DATA PORT ADDRESS TO REG. C
NXTSEC: INR	C		; INCREMENT TO CONTROL PORT
SASICK: INP	A		; FIRST CHECK FOR DRIVE READY
	STA	DISK$STAT	; STORE STATUS
	ANI	(CMND OR BUSY OR REQ OR POUT)
	CPI	(CMND OR BUSY OR REQ)  ; IF POUT DROPS,
	RZ			       ;  WE ARE INTO STATUS PHASE
	ANI	(CMND OR BUSY OR REQ)
	CPI	(BUSY OR REQ)	; WHEN CMND DROPS, SEEK IS COMPLETE, AND WE ARE
	JRNZ	SASICK		;  READY FOR DATA TRANSFER
	DCR	C		; DATA PORT ADDRESS TO REG. C
	MVI	B,128
HERE:	INIR			; CHANGED TO OUTIR FOR WRITE
	JR	NXTSEC

PAGE



;	CHECK STATUS OF READ OR WRITE
;
CHK$STAT:			; THIS ROUTINE CHECKS WHAT'S UP
	LXI	H,STAT		; STATUS BUFFER
	LDA	BASE$PORT
	MOV	D,A		; DATA PORT ADDRESS STORED IN REG. D
	INR	A
	MOV	E,A		; CONTROL PORT ADDRESS STORED IN REG. E
	JR	CHK01
CHKNXT: MOV	C,D		; INPUT FROM DATA PORT
	INP	A
	MOV	M,A		; SAVE IN MEMORY
CHK01:	MOV	C,E		; INPUT FROM CONTROL PORT
	INP	A
	ANI	(MSG OR REQ OR CMND OR POUT)
	CPI	(REQ OR CMND)
	JRZ	CHKNXT
	CPI	(MSG OR REQ OR CMND)
	JRNZ	CHK01
	MOV	C,D		; INPUT FROM DATA PORT
	INP	A		; GET FINAL BYTE
	MOV	A,M		; AND THROW IT AWAY, GET STATUS
	ANI	03		; EITHER BIT SET IS AN ERROR
	RET

****************************************************************

;
;	DATA BUFFERS AND STORAGE
;
LUNDATA:DB	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 ; 16 BYTES OF DATA FOR EACH
	DB	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 ; LOGICAL UNIT -- PULLED FROM
	DB	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 ; MAGIC SECTOR ON
	DB	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 ; INITIALIZATION
				     ; BYTE 1 : DRIVE/CONTROLLER TYPE
				     ; BYTE 2 : CONTROL BYTE
				     ; BYTES 3 - 10 : DRIVE CHARACTERISTIC DATA
				     ; BYTES 11 - 16 : ASSIGN DRIVE TYPE
				     ;				     COMMAND
LUN:	DB	0		; LOGICAL UNIT NUMBER
CMBFR:	DB	0,0,0,0,0,0	; COMMAND BUFFER
ICMND:	DB	INIT,0,0,0,0,0	; INITIALIZE DRIVE CHARACTERISTICS COMMAND
				;					 STRING
RCMND:	DB	RECAL,0,0,0,0,0 ; RESTORE COMMAND
DATAPTR:DW	0		; POINTER TO LUNDATA FOR THIS L.U.N.
SECPTR	DW	0		; POINTER TO CURRENT SECTOR TABLE ENTRY
SELERR	DB	0		; SELECT ERROR FLAG
FLAGS:	DB	0		; BIT 7 = INITIALIZATION FLAG,
				; BIT 6 = FLOPPY DISK FLAG
				; BIT 5 = REMOVABLE MEDIA FLAG
				; BIT 4 (SPARE)
				; BITS 0-3 = PARTITION NUMBER,
BASE$PORT:
	DB	0		; BASE PORT ADDRESS
INIT$FLAG:
	DB	1		; INITIALIZATION FLAG
STAT:	DB	0
;
; DEBLOCKING VARIABLES
;
PNDWRT	DB	0		; PENDING WRITE FLAG
RD$FLAG:DB	0		; FLAG FOR PRE-READ
HSTDSK	DB	0FFH
HSTTRK	DW	0
HSTSEC	DB	0
REQDSK	DB	0
REQTRK	DW	0
REQSEC	DB	0
BLCODE: DB	0		; SECTOR SIZE CODE (0=128,1=256,2=512,3=1024)
BLKSEC: DB	0		; LOCATION OF LOGICAL SECTOR WITHIN PHYSICAL
WRTYPE	DB	0
UNALLOC DB	0
URECORD DW	0
RELDSK	DB	0		; DISK NUMBER RELATIVE TO 0
CURDPB	DW	0		; ADDRESS OF CURRENT DISK PARAMETER BLOCK
BLKMSK	DB	0		; BLOCK MASK
OFFSET	DB	0

	REPT	(($+0FFH) AND 0FF00H)-$
	DB	0FFH
	ENDM
PAGE



MODLEN	EQU	$-MBASE

 DB 00100100B,10000000B,00000000B,00000000B,00000000B,00000000B,00000000B,00000001B
 DB 01010100B,00000001B,01010100B,00000001B,01010100B,00000001B,01010100B,00000001B
 DB 01010100B,00000001B,01010100B,00000001B,01010100B,00000001B,01010100B,00000001B
 DB 01010100B,00000000B,00000000B,00000000B,00000000B,00000000B,00000000B,00000000B
 DB 00000000B,00000000B,00000000B,00000000B,00000000B,00000000B,00000000B,00000000B
 DB 00000000B,00000000B,00000000B,00000000B,00000000B,00000000B,00000000B,00000000B
 DB 00000000B,00000000B,00000000B,00000000B,00000000B,00010010B,01001000B,10010000B
 DB 10010000B,00000100B,00010000B,10000000B,00000100B,10010000B,01000000B,01001001B
 DB 00100000B,01000000B,01000000B,01001000B,10000010B,00010000B,00000000B,00000000B
 DB 00001000B,10001000B,00010000B,00010000B,00000000B,10000000B,10001001B,00100010B
 DB 01001001B,00100100B,00100001B,00000001B,00000000B,01000000B,01000100B,00000010B
 DB 00000001B,00100000B,00000100B,00000001B,00100000B,10000000B,01000000B,00100010B
 DB 01000000B,00010001B,00000001B,00100001B,00100100B,10000100B,00100100B,00000000B
 DB 10010000B,00001000B,01001000B,00000010B,00100100B,00010000B,00010000B,10010000B
 DB 00100010B,00100000B,01000001B,00100000B,00000000B,01000100B,10010000B,00010001B
 DB 00100000B,10001000B,00010010B,00000000B,10000100B,01000010B,00010001B,00000001B
 DB 00000100B,00000000B,00010010B,01001001B,00100010B,00000000B,00001001B,00000001B
 DB 00100100B,01000000B,00100100B,10000000B,10001001B,00010010B,01001000B,10000000B
 DB 00100000B,00000001B,00000000B,00000000B,00100000B,00000100B,01000100B,00100100B
 DB 00100000B,00010001B,00010001B,00000000B,00100000B,00000000B,00000000B,00000000B
 DB 00000000B,00100100B,00000010B,00000000B,00000000B,00000000B,01000010B,00000000B
 DB 01001001B,00000010B,00000000B,00000000B,00010010B,00000000B,00000000B,00000000B
 DB 00000000B,00000000B,00000000B,00000000B,00000000B,00000000B,00000000B,00000000B
 DB 00000000B,00000000B,00000000B,00000000B,00000000B,00000000B,00000000B,00000000B

PAGE



********************************************************
** COMMON BUFFERS
********************************************************
	ORG	COMBUF
	DS	20
	DS	64
	DS	2
DIRBUF	DS	128
********************************************************

********************************************************
** BUFFERS
********************************************************
	ORG	BUFFER
HSTBUF: DS	512
CSV0:	DS	0
ALV0:	DS	256
ALV1:	DS	256
ALV2:	DS	256
ALV3:	DS	0
ALV4:	DS	0
ALV5:	DS	0
ALV6:	DS	0
ALV7:	DS	0
ALV8:	DS	0
**********************************************************
BUFLEN	EQU	$-BUFFER
	END


